;;;; packer.lisp -- a little language for packing binary structures
;;;
;;; This file implements something very similar to Perl's pack() function
;;; or Python's struct module--packing numeric datatypes and strings into
;;; a byte vector.  Due to the likelihood that this package will be used
;;; to interface with various binary formats, the only numeric datatypes
;;; supported are boring C datatypes--1, 2, 4, and 8-byte integers, along
;;; with single- and double-floats.
;;;
;;; Binary structures are specified with strings.  A specification contains
;;; several fields, where a field is [NUMBER] FIELD-CHAR.  FIELD-CHAR
;;; specifies a type of datum to be packed; NUMBER specifies the count of
;;; the particular datum to be packed and defaults to 1.  So, for instance,
;;; the specification "4f" indicates the binary structure would contain
;;; four single-float values.  Different FIELD-CHARs may be added with
;;; DEFINE-FORMAT-CHAR, below, although there are no hooks for a user to
;;; specify how certain objects should be packed.
;;;
;;; Packing is by default done in a big-endian fashion.  You may specify
;;; whether to do big-endian packing or little-endian packing with a
;;; #\> or #\<, respectively, at the beginning of the format string.
;;; e.g. the format string "<4s" would specify that four (SIGNED-BYTE 16)
;;; values should be packed with little-endian byte ordering.  Some
;;; format characters may pack data according to a particular byte
;;; ordering regardless of the byte-ordering specified by the format string.
;;; (Although this is not currently true of the predefined format
;;; characters).
;;;
;;; The only limitation you may run into is that I assume that SINGLE-FLOAT
;;; and DOUBLE-FLOAT are equivalent to the common C definitions of singles
;;; and doubles.  That is, SINGLE-FLOAT can be represented with 32 bits
;;; of data and DOUBLE-FLOAT can be represented with 64 bits of data.  If
;;; this assumption is not true for your Common Lisp implementation, you
;;; may encounter data corruption.
;;;
;;; FIXME: add a mechanism so that users can specify how to pack arbitrary
;;;   values into a binary structure.
;;; FIXME: this package should integrate well with BINARY-TYPES.  Figure
;;;   out how to make that so.

(in-package "BINARY-PACKER")


;;; errors than can be signaled

(define-condition binary-packer-error (simple-error)
  ()
  (:documentation "All errors signaled by BINARY-PACKER are of this type."))

(define-condition unknown-format-character-error (binary-packer-error)
  ((char :initarg :char :reader unknown-format-character))
  (:default-initargs :char nil)
  (:report (lambda (condition stream)
             (format stream "~?~@[: ~A~]"
                     (simple-condition-format-control condition)
                     (simple-condition-format-arguments condition)
                     (unknown-format-character condition)))))

(define-condition format-string-syntax-error (binary-packer-error)
  ((char :initarg :char :reader format-string-syntax-error-character))
  (:default-initargs :char nil)
  (:report (lambda (condition stream)
             (format stream "~?~@[: ~A~]"
                     (simple-condition-format-control condition)
                     (simple-condition-format-arguments condition)
                     (format-string-syntax-error-character condition))))
  (:documentation "Signaled when an error is encountered while parsing a format string."))

(define-condition unpacking-error (binary-packer-error)
  ()
  (:documentation "Signaled when too few values remain to be unpacked."))


;;; defining how to pack certain values in the abstract

(defparameter *pack-formats* nil
  "An alist mapping from format characters to format definitions.")

(defun find-format-definition-or-else (char)
  (let ((format-def (find char *pack-formats* :key #'fd-char :test #'char=)))
    (unless format-def
      (error 'unknown-format-character-error
             :format-control "Format character not defined"
             :char char))
    format-def))

(defclass format-definition ()
  ((char :initarg :char :reader fd-char)
   (bytes :initarg :bytes :initform nil :reader fd-bytes
          :documentation "How many bytes in a packed structure this value occupies.
May be NIL if value-dependent.")))

(defmacro define-format-char (format-char size)
  `(let* ((format-def (make-instance 'format-definition
                                     :char ,format-char
                                     :bytes ,size)))
    ;; FIXME: what to do if the format definition already exists?
    (push format-def *pack-formats*)))


;;; defining the packers themselves and how to pack simple numeric values

(defclass packer ()
  ((buffer :initform (make-array 64 :element-type '(unsigned-byte 8)
                                :initial-element 0)
          :reader buffer)
   (index :initform 0 :accessor index)))
(defclass little-endian-packer (packer) ())
(defclass big-endian-packer (packer) ())

(defclass unpacker ()
  ((buffer :initarg :buffer :reader buffer :type (vector (unsigned-byte 8)))
   (index :initarg :index :initform 0 :accessor index)))
(defclass little-endian-unpacker (unpacker) ())
(defclass big-endian-unpacker (unpacker) ())

(defun ensure-space-for-packing (packer n-bytes)
  (with-slots (buffer index) packer
    (let ((length (length buffer)))
      (when (>= (+ index n-bytes) length)
        (let ((new-buffer (make-array (max (+ index n-bytes) (* length 2))
                                      :element-type '(unsigned-byte 8))))
          (replace new-buffer buffer)
          (setf buffer new-buffer)))
      (values))))

(defun basic-numeric-packer (packer pack-fun transform-fun objects count)
  (loop for i from 0 below count
        for (value . rest) on objects
        do (funcall pack-fun packer (funcall transform-fun value))
        finally (return rest)))

(defun ensure-values-for-unpacking (unpacker n-bytes)
  (with-slots (buffer index) unpacker
    (unless (<= (+ index n-bytes) (length buffer))
      (error 'unpacker-error
             :format-control "Too few values left in the unpacker buffer"))
    (values)))

(defun basic-numeric-unpacker (unpacker unpack-fun inverter-fun count)
  (loop for i from 0 below count
        collect (funcall inverter-fun (funcall unpack-fun unpacker))))

(eval-when (:compile-toplevel)
(defun ldb-forms-for-packing (bits &optional big-endian)
  (loop for pos from 0 below bits by 8
        collect `(ldb (byte 8 ,pos) value) into loads
        finally (return (if big-endian
                            (reverse loads)
                            loads))))

(defun dpb-form-for-unpacking (bits &optional big-endian)
  (let ((n-bytes (truncate bits 8)))
    (do ((i 0 (1+ i))
         (form 0))
        ((>= i n-bytes) form)
      (setf form `(dpb (aref buffer (+ index ,(if big-endian
                                                  (- n-bytes i 1)
                                                  i)))
                   (byte 8 ,(* i 8))
                   ,form)))))
) ; EVAL-WHEN

(macrolet ((define-pack-method (name bits n-bytes &key big-endian signed)
               (let ((endian-name
                      (intern (format nil
                                      "~A-~:[L~;B~]E" name big-endian)))
                     (type (if signed
                               `(signed-byte ,bits)
                               `(unsigned-byte ,bits))))
                 `(defun ,endian-name (packer value)
                   (declare (type ,type value))
                   (ensure-space-for-packing packer ,n-bytes)
                   (let ((buffer (buffer packer))
                         (index (index packer)))
                     (declare (type (simple-array (unsigned-byte 8) (*)) buffer))
                     (declare (type (integer 0 ,(- array-total-size-limit n-bytes)) index))
                     ,(loop for form in (ldb-forms-for-packing bits big-endian)
                            for i from 0
                            collect `(setf (aref buffer (+ index ,i))
                                      ,form) into setfs
                            finally (return `(progn ,@setfs)))
                     (setf (index packer) (+ index ,n-bytes))
                     (values)))))
           (define-unpack-method (name bits n-bytes &key big-endian signed)
               (let ((endian-name
                      (intern (format nil
                                      "~A-~:[L~;B~]E" name big-endian))))
                 `(defun ,endian-name (unpacker)
                   (ensure-values-for-unpacking unpacker ,n-bytes)
                   (let* ((buffer (buffer unpacker))
                          (index (index unpacker))
                          (value ,(dpb-form-for-unpacking bits big-endian)))
                     (declare (type (vector (unsigned-byte 8)) buffer))
                     (declare (type (integer 0 ,(- array-total-size-limit n-bytes)) index))
                     (declare (type (unsigned-byte ,bits) value))
                     (prog1 ,(if signed
                                 `(when (logbitp (1- ,bits) value)
                                   (dpb value (byte ,bits 0) -1))
                                 'value)
                       (setf (index unpacker) (+ index ,n-bytes)))))))
           (defpackfuns (bits)
               (let ((pack-ub (format nil "PACK-UB~A" bits))
                     (pack-sb (format nil "PACK-SB~A" bits))
                     (unpack-ub (format nil "UNPACK-UB~A" bits))
                     (unpack-sb (format nil "UNPACK-SB~A" bits))
                     (n-bytes (truncate bits 8)))
               `(progn
                 (define-pack-method ,pack-ub ,bits ,n-bytes :big-endian nil)
                 (define-pack-method ,pack-ub ,bits ,n-bytes :big-endian t)
                 (define-pack-method ,pack-sb ,bits ,n-bytes
                                     :big-endian nil :signed t)
                 (define-pack-method ,pack-sb ,bits ,n-bytes
                                     :big-endian t :signed t)
                 (define-unpack-method ,unpack-ub ,bits ,n-bytes
                                       :big-endian nil)
                 (define-unpack-method ,unpack-ub ,bits ,n-bytes
                                       :big-endian t)
                 (define-unpack-method ,unpack-sb ,bits ,n-bytes
                                       :big-endian nil :signed t)
                 (define-unpack-method ,unpack-sb ,bits ,n-bytes
                                       :big-endian t :signed t)))))
  (defpackfuns 8)
  (defpackfuns 16)
  (defpackfuns 32)
  (defpackfuns 64))


;;; packings and unpackings for non-simple values

(defun single-float-bits (f)
  (multiple-value-bind (mantissa exponent sign) (integer-decode-float f)
    (let ((actual-sign (if (= sign 1) 0 1)))
      (dpb actual-sign (byte 1 31)
           (dpb exponent (byte 7 24) mantissa)))))

(defun double-float-bits (f)
  (multiple-value-bind (mantissa exponent sign) (integer-decode-float f)
    (let ((actual-sign (if (= sign 1) 0 1)))
      (dpb actual-sign (byte 1 63)
           (dpb exponent (byte 10 53) mantissa)))))

(defun make-single-float-from-bits (bits)
  (let ((signif (ldb (byte 24 0) bits))
        (exponent (ldb (byte 7 24) bits))
        (sign (if (zerop (ldb (byte 1 31) bits))
                  1
                  -1)))
    (* (scale-float (float signif 1.0f0) exponent) sign)))

(defun make-double-float-from-bits (bits)
  (let ((signif (ldb (byte 53 0) bits))
        (exponent (ldb (byte 10 53) bits))
        (sign (if (zerop (ldb (byte 1 63) bits))
                  1
                  -1)))
    (* (scale-float (float signif 1.0d0) exponent) sign)))


;;; generic functions which participate in packing and unpacking

(defgeneric pack-objects (packer objects char count)
  (:documentation "Pack COUNT values from OBJECTS into PACKER based on
CHAR.  Return the unpacked objects.  Some methods of this function may
modify PACKER but consume fewer objects than COUNT would indicate."))

;;; FIXME: this will actually hork for a combination of large
;;; (~ ARRAY-TOTAL-SIZE-LIMIT) strings and/or large 
;;; (~ ARRAY-TOTAL-SIZE-LIMIT) amounts of data already packed into the
;;; packer.  This condition should be easily detectable, though.
(defmethod pack-objects (packer objects (char (eql #\a)) count)
  (let* ((string (car objects))
         (length (length string)))
    (ensure-space-for-packing packer count)
    (let ((buffer (buffer packer))
          (index (index packer)))
      (declare (type (simple-array (unsigned-byte 8) (*)) buffer))
      (do ((i 0 (1+ i)))
          ((>= i count)
           (incf (index packer) count)
           (cdr objects))
        (let ((value (if (< i length)
                         (char-code (char string i))
                         0)))           ; pad with NULL
          (setf (aref buffer (+ index i)) value))))))

(defmethod pack-objects (packer objects (char (eql #\x)) count)
  (ensure-space-for-packing packer count)
  (dotimes (i count objects)
    ;; blech, FIXME
    (pack-ub8-be packer 0)))

(defgeneric unpack-objects (unpacker char count)
  (:documentation "Unpack values from UNPACKER based on CHAR.
Return the unpacked objects as a list, even if there is only one.
COUNT is an upper bound on how many objects the resulting list will
contain, although some format characters may interpret COUNT in a 
way other than the number of objects to be returned."))

(defmethod unpack-objects (unpacker (char (eql #\a)) count)
  (ensure-values-for-unpacking unpacker count)
  (let ((string (make-string count :element-type 'base-char))
        (buffer (buffer unpacker))
        (index (index unpacker)))
    (dotimes (i count (progn
                        (incf (index unpacker) count)
                        string))
      (setf (aref string i) (code-char (aref buffer (+ index i)))))))

(defmethod unpack-objects (unpacker (char (eql #\x)) count)
  (ensure-values-for-unpacking unpacker count)
  ;; unpacking nulls is really easy
  (incf (index unpacker) count)
  ;; return nothing
  nil)

(defun align (offset size)
  (declare (type fixnum offset size))
  (+ offset (let ((x (rem offset size)))
              (if (zerop x) x (- size x)))))

(defmacro define-numeric-pack-routines (format-char bytes type
                                                    &key signed
                                                    (transformer 'identity)
                                                    (inverter 'identity))
  (declare (ignore type))
  (loop for big-endian in (list t nil)
        collect
        (let ((packer-class (intern (format nil "~:[LITTLE~;BIG~]-ENDIAN-PACKER"
                                            big-endian)))
              (packer (intern (format nil "PACK-~:[U~;S~]B~A-~:[L~;B~]E"
                                signed (* bytes 8) big-endian))))
          `(defmethod pack-objects ((packer ,packer-class) objects (char (eql ,format-char)) count)
            (setf (index packer) (align (index packer) ,bytes))
            (basic-numeric-packer packer #',packer #',transformer objects count)))
        into functions
        collect
        (let ((unpacker-class (intern (format nil "~:[LITTLE~;BIG~]-ENDIAN-UNPACKER"
                                            big-endian)))
              (unpacker (intern (format nil "UNPACK-~:[U~;S~]B~A-~:[L~;B~]E"
                                        signed (* bytes 8) big-endian))))
          `(defmethod unpack-objects ((unpacker ,unpacker-class) (char (eql ,format-char)) count)
            (setf (index unpacker) (align (index unpacker) ,bytes))
            (basic-numeric-unpacker unpacker #',unpacker #',inverter count)))
        into functions
        finally (return `(progn
                          (define-format-char ,format-char ,bytes)
                          ,@functions))))


;;; predefined format characters, mostly taken from Python, with some
;;; inspiration from Perl as well.

(define-numeric-pack-routines #\c 1 base-char
  :transformer char-code
  :inverter code-char)
(define-numeric-pack-routines #\b 1 (signed-byte 8) :signed t)
(define-numeric-pack-routines #\B 1 (unsigned-byte 8))
(define-numeric-pack-routines #\s 2 (signed-byte 16) :signed t)
(define-numeric-pack-routines #\S 2 (unsigned-byte 16))
(define-numeric-pack-routines #\i 4 (signed-byte 32) :signed t)
(define-numeric-pack-routines #\I 4 (unsigned-byte 32))
(define-numeric-pack-routines #\l 8 (signed-byte 64) :signed t)
(define-numeric-pack-routines #\L 8 (unsigned-byte 64))
(define-numeric-pack-routines #\f 4 single-float
  :transformer single-float-bits
  :inverter make-single-float-from-bits)
(define-numeric-pack-routines #\d 8 double-float
  :transformer double-float-bits
  :inverter make-double-float-from-bits)

(define-format-char #\x 1)
(define-format-char #\a 1)

(defun check-format-type (format-char object)
  (assert (typep object (cdr (assoc format-char *pack-types* :test #'char=)))))


;;; parsing of packing format strings

(defclass field ()
  ((format-char :reader format-char :initarg :format-char :type character)
   (count :reader repeat-count :initarg :count :initform 1 :type fixnum)))

(defmethod print-object ((object field) stream)
  (print-unreadable-object (object stream)
    (if (= 1 (repeat-count object))
        (format stream "Field ~A" (format-char object))
        (format stream "Field ~A~A" (repeat-count object)
                (format-char object)))))

(defun field-element-size (field)
  (object-size (format-char field)))

(defun parse-format-string (format-string)
  (declare (type simple-string format-string))
  (let ((index 0)
        (length (length format-string))
        (count nil)
        (fields nil)
        (big-endian t))
    (declare (type fixnum index length))
    (flet ((char-to-num (char)
             (- (char-code char) (char-code #\0)))
           (valid-format-char-p (char)
             (find char *pack-formats* :key #'fd-char :test #'char=)))
      (tagbody
       MAYBE-ENDIANIZE
         (when (= index length)
           (go END))
         (let ((char (char format-string index)))
           (cond
             ((char= char #\<)
              (setf big-endian nil)
              (incf index))
             ((char= char #\>)
              (setf big-endian t)
              (incf index)))
           (go START))
       START
         (when (= index length)
           (go END))
         (let ((char (char format-string index)))
           (cond
             ((digit-char-p char)
              (setf count (char-to-num char))
              (incf index)
              (go NUMBER))
             ((char= char #\Space) (incf index) (go START))
             ((valid-format-char-p char)
              (push (make-instance 'field :format-char char) fields)
              (incf index)
              (go START))
             (t (error 'format-string-syntax-error
                       :format-control "Invalid character encountered at position ~A"
                       :format-arguments (list index)
                       :char char))))
       NUMBER
         (when (= index length)
           (error 'format-string-syntax-error
                  :format-control "Unexpected end-of-string while parsing format string ~S"
                  :format-arguments (list format-string)))
         (let ((char (char format-string index)))
           (cond
             ((digit-char-p char)
              (setf count (+ (char-to-num char) (* 10 count)))
              (incf index)
              (go NUMBER))
             (t
              (go FORMAT-CHAR))))
       FORMAT-CHAR
         (when (= index length)
           (error 'format-string-syntax-error
                  :format-control "Unexpected end-of-string while parsing format string ~S"
                  :format-arguments (list format-string)))
         (let ((char (char format-string index)))
           (cond
             ((valid-format-char-p char)
              (push (make-instance 'field :format-char char :count count)
                    fields)
              (incf index)
              (setf count nil)
              (go START))
             (t (error 'unknown-format-character-error
                       :format-control "Invalid format character encountered at position ~A"
                       :format-arguments (list index)
                       :char char))))
       END
         (return-from parse-format-string
           (values (nreverse fields) big-endian))))))

(defun field-list-size (field-list)
  (let ((size 0))
    (dolist (field field-list size)
      (let* ((format-def (find-format-definition-or-else (format-char field)))
             (field-size (fd-bytes format-def)))
        (unless size
          (return-from field-list-size nil))
        (setf size (+ (align size field-size)
                      (* (repeat-count field) field-size)))))))

(defun pack-size (format-string)
  "Return the size of the (UNSIGNED-BYTE 8) vector determined by FORMAT-STRING.
May return NIL if the size of a field in FORMAT-STRING is value-dependent."
  (field-list-size (parse-format-string format-string)))


;;; doing the actual packing and unpacking

(defun pack-field (packer field objects)
  (pack-objects packer objects (format-char field) (repeat-count field)))

(defun pack (format-string &rest objects)
  "Pack OBJECTS into a (VECTOR (UNSIGNED-BYTE 8)) according to FORMAT-STRING.
Return the packed vector."
  (multiple-value-bind (fields big-endian-p) (parse-format-string format-string)
    (let ((packer (if big-endian-p
                      (make-instance 'big-endian-packer)
                      (make-instance 'little-endian-packer))))
      (loop while objects
        for field in fields
        do (setf objects (pack-field packer field objects))
        finally (let* ((buffer (buffer packer))
                       (index (index packer))
                       (retbuf (make-array index :element-type '(unsigned-byte 8)
                                           :initial-element 0)))
                  (declare (type (simple-array (unsigned-byte 8) (*)) buffer retbuf))
                  (return (replace retbuf buffer)))))))

(defun unpack-field (unpacker field)
  (unpack-objects unpacker (format-char field) (repeat-count field)))

(defun unpack (format-string buffer &key (start 0) end)
  "Unpack objects from the (VECTOR (UNSIGNED-BYTE 8)) BUFFER according to
FORMAT-STRING.  Return the list of unpacked objects."
  (let* ((end (or end (length buffer)))
         ;; we could make this ever so slightly faster by checking to
         ;; see if (AND (ZEROP START) (= END (LENGTH BUFFER))) and not
         ;; creating the displaced array, but it's probably not worth it.
         (buffer (make-array (- end start)
                             :displaced-to buffer
                             :displaced-index-offset start)))
    (multiple-value-bind (fields big-endian-p) (parse-format-string format-string)
      (let ((unpacker (if big-endian-p
                          (make-instance 'big-endian-unpacker :buffer buffer)
                          (make-instance 'little-endian-unpacker :buffer buffer)))
            (objects nil))
        (dolist (field fields objects)
          (let ((unpacked-objects (unpack-field unpacker field)))
            (setf objects (nconc objects unpacked-objects))))))))
